#!/usr/bin/env python

########################################################
# 
# Python version of the original MATLAB reweight 
# script by P. Tiwary
# https://sites.google.com/site/pratyushtiwary/scripts
# Tiwary and Parrinello JPCB 2014
#
#
# L. Sutto
# l.sutto@ucl.ac.uk                    v1.1 - 16/12/2014
########################################################

########################################################
# INPUT SETTINGS
########################################################
# Well-Tempered Metadynamics or not
tempered = True

# biasfactor for Well-Tempered
gamma = 30

# kT in your energy units (kJ or kcal)
kT = 2.4943

# number of FES file generated with sum_hills stride option
# the more the better
numdat = 350

# input FES file prefix
fesfilename ='fes_'

# column in FES file corresponding to the Free Energy
# NB: the first column is 0
col_fes = 1

# name of the file containing both the original CVs, the reweight CVs
# and the bias
datafile = 'COLVAR'

# list with column numbers of your datafile containing the bias
# and any external bias/restraint/walls
# NB: the first column is 0
col_bias = [ 3, 7, 8, 9, 10, 11, 12 ] 
#col_bias = [ 3 ]

# list with column numbers of your datafile containing the CVs to be reweighted
# NB: the first column is 0
col_rewt = [ 1 ]

# If calc_s_range = True, then find the minimum and maximum from the datafile, otherwise use these ones
calc_s_range = False
s_min = [ -2.9 ]
s_max = [ 2.3]

# grid size for the reweighted FES
ngrid = [ 521 ]

# output FES filenames for 2 different formats
out_fes = 'fes_rew.dat'
out_fes_xy = 'fesxy_rew.dat'

# print some output while running
verbose = True
########################################################


########################################################
# FIRST PART: calculate c(t)
# This part is independent of the number of CVs being biased
# Physically, c(t) represents an estimate of the reversible 
# work performed on the system until time t
########################################################

import numpy as np
from math import log, exp, ceil

if verbose: print "Reading FES files.."

# calculates exp(beta c(t)), using eq. 12 and eq. 3 in the JPCB
ebetac = []
for i in range(numdat):

    if verbose and i%(numdat/10)==0: print "%d of %d (%.0f%%) done" % (i,numdat,(i*100./numdat))

    ########################################
    # set appropriate format for FES file names, NB: i starts from 0
    fname = '%s%d.dat' % (fesfilename,i)
#    fname = '%s.%d' % (fesfilename,i+1)
    ########################################

    data = np.loadtxt(fname)
    s1, s2 = 0., 0.
    if tempered:
        for p in data:
            s1 += exp(-p[col_fes]/kT)
            s2 += exp(-p[col_fes]/kT/gamma)
    else:
        for p in data:
            s1 += exp(-p[col_fes]/kT)
        s2 = len(data)
    ebetac += s1 / s2,

# this would be c(t)
# coft = [ kT*log(x) for x in ebetac ]

########################################################
# SECOND PART: Boltzmann-like sampling for reweighting
########################################################

if verbose: print "Calculating reweighted FES.."

# NB: loadtxt takes care of ignoring comment lines starting with '#'
colvar = np.loadtxt(datafile)

# number of reweighting CVs
numrewt = len(col_rewt)

# find min and max of reweighting CVs
if calc_s_range:
    s_min = [ 9e99 ] * numrewt
    s_max = [ -9e99 ] * numrewt

numcolv = 0
for row in colvar:
    numcolv += 1
    if calc_s_range:
       for i in range(numrewt):
           col = col_rewt[i]
           val = row[col]
           if val < s_min[i] : s_min[i] = val 
           if val > s_max[i] : s_max[i] = val

if verbose: 
    for i in range(numrewt):
        print "CV[%d] range: %f ; %f" % (i,s_min[i],s_max[i])

# build the new grid for the reweighted FES
s_grid = [[ ]] * numrewt
ds = [ 0.0 ] * numrewt
for i in range(numrewt):
    ds[i] = (s_max[i] - s_min[i])/(ngrid[i]-1)
    s_grid[i] = [ s_min[i] + n*ds[i] for n in range(ngrid[i]) ]
    if verbose: print "Grid size CV[%d]=%f" % (i,ds[i])

# initialize rectangular  array numrewt-dimensional
fes = np.zeros( [ngrid[i] for i in range(numrewt)] )

# go through the CV(t) trajectory

print numcolv
denom = 0.
i = 0
for row in colvar:
    i += 1

    # build the array of grid indeces locs corresponding to the point closest to current point
    locs = [[ ]] * numrewt
    for j in range(numrewt):
        col = col_rewt[j]
        val = row[col]
        locs[j] = [int((val-s_min[j])/ds[j])]
         
    #find closest c(t) for this point of time
    indx = int(ceil(float(i)/numcolv*numdat))-1
    bias = sum([row[j] for j in col_bias])
    ebias = exp(bias/kT)/ebetac[indx]
    fes[locs] += ebias
    denom += ebias

# ignore warnings about log(0) and /0
np.seterr(all='ignore')

fes /= denom
fes = -kT*np.log(fes)
# set the minimum to 0
fes -= np.min(fes)


########################################################
# OUTPUT RESULTS ON FILES
########################################################

if verbose: print "Saving results on %s and %s" % (out_fes,out_fes_xy)

# save the FES in the format: FES(x,y) (one increment of y per row)
np.savetxt(out_fes, fes, fmt='%e', delimiter=' ')

# print the FES in the format: x,y,FES(x,y) if 2D or x,FES(x) if 1D
with open(out_fes_xy, 'w') as f:
    if numrewt==2:
        for ny,y in enumerate(s_grid[1]):
            for nx,x in enumerate(s_grid[0]):
                f.write('%f %f %f\n' % (x,y,fes[nx][ny]))
            f.write('\n')
    elif numrewt==1:
        for nx,x in enumerate(s_grid[0]):
            f.write('%f %f\n' % (x,fes[nx]))
f.close()

